import dayjs from 'dayjs';
import { get, reduce } from 'lodash-es';
import { v4 } from 'uuid';

export enum FrequencyShifting {
  /**
   * Change frequency undergoes a linear transformation from previous pulse to current one.
   */
  Linear = 0,
  /**
   * Change frequency undergoes a quadratic transformation from previous pulse to current one.
   */
  Quadratic = 1,
  /**
   * Change frequency undergoes a cubic transformation from previous pulse to current one.
   */
  Cubic = 2,
  /**
   * Change frequency with quick fade in and fade out.
   */
  Ease = 3,
  /**
   * Change frequency with spiking within range from previous pulse and current one.
   */
  Pulsating = 4,
  /**
   * Based on frequency of previous pulse, take the twice of frequency of current pulse as the peak frequency. Follow the maximum frequency limitation.
   */
  Spiking = 5,
  /**
   * Randomize frequency within range from previous frequency and current one.
   */
  Randomize = 6,
  /**
   * Randomize frequency within minium and maximum frequency.
   */
  Maniac = 7,
  /**
   * Synchronize changes of frequency with pulse width changes.
   */
  Synchronized = 8,
}

export interface ControlPoint {
  x: number;
  y: number;
}

export interface Pulse {
  order: number;
  id: number;
  offset: number;
  width: number;
  maniac: boolean;
  frequency: number;
  frequencyShifting: FrequencyShifting;
  controlPoint1: ControlPoint;
  controlPoint2: ControlPoint;
}
export function createPulse(order: number, width: number, frequency: number) {
  return {
    id: v4(),
    order,
    offset: 0,
    width,
    maniac: false,
    frequency,
    frequencyShifting: FrequencyShifting.Linear,
    controlPoint1: { x: 0, y: 0 },
    controlPoint2: { x: 0, y: 0 },
  } as Pulse;
}

export interface Pattern {
  id: string;
  name: string;
  createdAt: number;
  lastModifiedAt: number | null;
  smoothRepeat: boolean;
  pulses: Pulse[];
}
export function createPattern() {
  return {
    id: v4(),
    name: '',
    createdAt: dayjs().valueOf(),
    lastModifiedAt: null,
    smoothRepeat: true,
    pulses: [],
  } as Pattern;
}

export function movePulseUp(pattern: Pattern, pulseId: string, step: number) {
  const index = pattern.pulses.findIndex((pulse) => pulse.id === pulseId);
  if (index === -1 || index - step < 0) return;

  const targetIndex = index - step;
  const targetPulse = pattern.pulses[targetIndex];
  const currentPulse = pattern.pulses[index];

  // Swap the pulses
  pattern.pulses[targetIndex] = currentPulse;
  pattern.pulses[index] = targetPulse;

  // If the target pulse's order is 1, swap their offsets
  if (targetPulse.order === 1) {
    const tempOffset = currentPulse.offset;
    currentPulse.offset = targetPulse.offset;
    targetPulse.offset = tempOffset;
  }

  // Swap their order
  const tempOrder = currentPulse.order;
  currentPulse.order = targetPulse.order;
  targetPulse.order = tempOrder;

  // Sort pulses by order
  pattern.pulses.sort((a, b) => a.order - b.order);
}

export function movePulseDown(pattern: Pattern, pulseId: string, step: number) {
  const index = pattern.pulses.findIndex((pulse) => pulse.id === pulseId);
  if (index === -1 || index + step >= pattern.pulses.length) return;

  const targetIndex = index + step;
  const targetPulse = pattern.pulses[targetIndex];
  const currentPulse = pattern.pulses[index];

  // Swap the pulses
  pattern.pulses[targetIndex] = currentPulse;
  pattern.pulses[index] = targetPulse;

  // If the current pulse's order is 1, swap their offsets
  if (currentPulse.order === 1) {
    const tempOffset = currentPulse.offset;
    currentPulse.offset = targetPulse.offset;
    targetPulse.offset = tempOffset;
  }

  // Swap their order
  const tempOrder = currentPulse.order;
  currentPulse.order = targetPulse.order;
  targetPulse.order = tempOrder;

  // Sort pulses by order
  pattern.pulses.sort((a, b) => a.order - b.order);
}

export function addPulse(pattern: Pattern): Pulse {
  const maxOrder = reduce(pattern.pulses, (former, pulse) => Math.max(former, pulse.order), 0);
  const newPulse = createPulse(
    maxOrder + 1,
    pattern.smoothRepeat
      ? get(pattern.pulses, '[0].width', 0)
      : get(pattern.pulses, '[-1].width', 0),
    pattern.smoothRepeat
      ? get(pattern.pulses, '[0].frequency', 1)
      : get(pattern.pulses, '[-1].frequency', 1),
  );
  pattern.pulses.push(newPulse);
  return newPulse;
}

export function updatePulse(pattern: Pattern, pulseId: string, pulse: Pulse) {
  const index = pattern.pulses.findIndex((p) => p.id === pulseId);
  if (index !== -1) {
    const { id, order, ...rest } = pulse;
    pattern.pulses[index] = { ...pattern.pulses[index], ...rest };
  }
}

export function deletePulse(pattern: Pattern, pulseId: string) {
  pattern.pulses = pattern.pulses.filter((pulse) => pulse.id !== pulseId);
  pattern.pulses.sort((a, b) => a.order - b.order);
  pattern.pulses.forEach((pulse, index) => {
    pulse.order = index + 1;
  });
}

export function durationRemains(pattern: Pattern): number {
  return 20 * 1000 - totalDuration(pattern) - (pattern.smoothRepeat ? 100 : 0);
}

export function totalDuration(pattern: Pattern): number {
  return reduce(
    pattern.pulses,
    (former, pulse) => former + pulse.offset,
    pattern.smoothRepeat && pattern.pulses.length > 1 ? 100 : 0,
  );
}
